@@ -147,6 +147,14 @@ fn hash<const K: usize, const V: usize>(k: &[u8; K], v: &Option<[u8; V]>) -> u32
     hasher.finalize() ^ 0xFF
 }
 
+#[inline]
+fn hash_batch_len(len: usize) -> u32 {
+    let mut hasher = crc32fast::Hasher::new();
+    hasher.update(&(len as u64).to_le_bytes());
+
+    hasher.finalize() ^ 0xFF
+}
+
 enum WorkerMessage {
     NewSST { id: u64, sst_sz: u64, db_sz: u64 },
     Stop(mpsc::Sender<()>),
@@ -178,7 +186,11 @@ impl<const K: usize, const V: usize> Worker<K, V> {
             // only compact one run at a time before checking
             // for new messages.
             if let Err(e) = self.sstable_maintenance() {
-                log::error!("error while compacting sstables in the background: {:?}", e);
+                log::error!(
+                    "error while compacting sstables \
+                    in the background: {:?}",
+                    e
+                );
             }
         }
         log::info!("tiny-lsm compaction worker quitting");
@@ -450,6 +462,9 @@ pub struct Lsm<const K: usize, const V: usize> {
     worker_outbox: mpsc::Sender<WorkerMessage>,
     next_sstable_id: u64,
     dirty_bytes: usize,
+    #[cfg(test)]
+    pub log: tearable::Tearable<fs::File>,
+    #[cfg(not(test))]
     log: BufWriter<fs::File>,
     path: PathBuf,
     config: Config,
@@ -542,23 +557,54 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         let mut reader = BufReader::new(log);
 
-        let mut buf = vec![0; 5 + K + V];
+        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
+        let mut buf = vec![0; 5 + min_tuple_sz];
 
         let mut memtable = BTreeMap::new();
         let mut recovered = 0;
+        let mut write_batch = None;
+        let mut wb_remaining = 0;
+        let mut wb_recovered = 0;
         while let Ok(()) = reader.read_exact(&mut buf) {
-            let crc_expected: u32 = u32::from_le_bytes(buf[0..4].try_into().unwrap());
-            let d: bool = match buf[4] {
-                0 => false,
-                1 => true,
-                _ => {
+            macro_rules! rewind {
+                () => {
                     // need to back up a few bytes to chop off the torn log
+                    log::warn!("detected log tear. chopping length to {}", recovered);
                     let log_file = reader.get_mut();
                     log_file.seek(io::SeekFrom::Start(recovered))?;
                     log_file.set_len(recovered)?;
                     log_file.sync_all()?;
                     fs::File::open(path.join(SSTABLE_DIR))?.sync_all()?;
                     break;
+                };
+            }
+            let crc_expected: u32 = u32::from_le_bytes(buf[0..4].try_into().unwrap());
+            let d: bool = match buf[4] {
+                0 => false,
+                1 => true,
+                2 if write_batch.is_none() => {
+                    // begin batch
+                    let batch_sz_buf: [u8; 8] = buf[5..5 + 8].try_into().unwrap();
+                    let batch_sz: u64 = u64::from_le_bytes(batch_sz_buf);
+
+                    if batch_sz > usize::MAX as u64 {
+                        return Err(io::Error::new(
+                            io::ErrorKind::InvalidInput,
+                            "recovering a batch size over usize::MAX is not supported",
+                        ));
+                    }
+
+                    let crc_actual = hash_batch_len(usize::try_from(batch_sz).unwrap());
+                    if crc_expected != crc_actual {
+                        rewind!();
+                    }
+
+                    write_batch = Some(Vec::with_capacity(batch_sz as usize));
+                    wb_remaining = crc_expected as usize;
+                    continue;
+                }
+                _ => {
+                    rewind!();
                 }
             };
             let k: [u8; K] = buf[5..K + 5].try_into().unwrap();
@@ -570,23 +616,42 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
             let crc_actual: u32 = hash(&k, &v);
 
             if crc_expected != crc_actual {
-                // need to back up a few bytes to chop off the torn log
-                let log_file = reader.get_mut();
-                log_file.seek(io::SeekFrom::Start(recovered))?;
-                log_file.set_len(recovered)?;
-                log_file.sync_all()?;
-                fs::File::open(path.join(SSTABLE_DIR))?.sync_all()?;
-                break;
+                rewind!();
             }
 
-            recovered += buf.len() as u64;
+            if let Some(ref mut wb) = write_batch {
+                wb.push((k, v));
+                wb_remaining -= 1;
+                wb_recovered += buf.len() as u64;
 
-            memtable.insert(k, v);
+                // apply the write batch all at once
+                // or never at all
+                if wb_remaining == 0 {
+                    for (k, v) in std::mem::take(wb) {
+                        memtable.insert(k, v);
 
-            if let Some(v) = v {
-                db.insert(k, v);
+                        if let Some(v) = v {
+                            db.insert(k, v);
+                        } else {
+                            db.remove(&k);
+                        }
+                    }
+                    recovered += wb_recovered;
+                    wb_recovered = 0;
+                }
             } else {
-                db.remove(&k);
+                assert_eq!(0, wb_remaining);
+                assert_eq!(0, wb_recovered);
+
+                memtable.insert(k, v);
+
+                if let Some(v) = v {
+                    db.insert(k, v);
+                } else {
+                    db.remove(&k);
+                }
+
+                recovered += buf.len() as u64;
             }
         }
 
@@ -612,7 +677,10 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
         for _ in hb_rx {}
 
         let lsm = Lsm {
+            #[cfg(not(test))]
             log: BufWriter::with_capacity(config.log_bufwriter_size, reader.into_inner()),
+            #[cfg(test)]
+            log: tearable::Tearable::new(reader.into_inner()),
             path: path.into(),
             next_sstable_id: max_sstable_id.unwrap_or(0) + 1,
             dirty_bytes: recovered as usize,
@@ -645,6 +713,11 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
     pub fn insert(&mut self, k: [u8; K], v: [u8; V]) -> Result<Option<[u8; V]>> {
         assert_ne!([k[0], v[0]], [255, 254]);
         self.log_mutation(k, Some(v))?;
+
+        if self.dirty_bytes > self.config.max_log_length {
+            self.flush()?;
+        }
+
         Ok(self.db.insert(k, v))
     }
 
@@ -657,9 +730,52 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
     /// durable, use the `Lsm::flush` method below.
     pub fn remove(&mut self, k: &[u8; K]) -> Result<Option<[u8; V]>> {
         self.log_mutation(*k, None)?;
+
+        if self.dirty_bytes > self.config.max_log_length {
+            self.flush()?;
+        }
+
         Ok(self.db.remove(k))
     }
 
+    /// Apply a set of updates to the `Lsm` and
+    /// log them to disk in a way that will
+    /// be recovered only if every update is
+    /// present.
+    pub fn write_batch(&mut self, write_batch: &[([u8; K], Option<[u8; V]>)]) -> Result<()> {
+        let batch_len: [u8; 8] = (write_batch.len() as u64).to_le_bytes();
+        let crc = hash_batch_len(write_batch.len());
+
+        self.log.write_all(&crc.to_le_bytes())?;
+        self.log.write_all(&[2_u8])?;
+        self.log.write_all(&batch_len)?;
+
+        // the zero pad is necessary because every log
+        // entry must have the same length, whether
+        // it's a batch size or actual kv tuple.
+        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
+        let pad_sz = min_tuple_sz - (K + V);
+        let pad = [0; std::mem::size_of::<u64>()];
+        self.log.write_all(&pad[..pad_sz])?;
+
+        for (k, v_opt) in write_batch {
+            if let Some(v) = v_opt {
+                self.db.insert(*k, *v);
+            } else {
+                self.db.remove(k);
+            }
+
+            self.log_mutation(*k, *v_opt)?;
+            self.memtable.insert(*k, *v_opt);
+        }
+
+        if self.dirty_bytes > self.config.max_log_length {
+            self.flush()?;
+        }
+
+        Ok(())
+    }
+
     fn log_mutation(&mut self, k: [u8; K], v: Option<[u8; V]>) -> Result<()> {
         let crc: u32 = hash(&k, &v);
         self.log.write_all(&crc.to_le_bytes())?;
@@ -672,7 +788,15 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
             self.log.write_all(&[0; V])?;
         };
 
-        let logged_bytes = 4 + 1 + K + V;
+        // the zero pad is necessary because every log
+        // entry must have the same length, whether
+        // it's a batch size or actual kv tuple.
+        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
+        let pad_sz = min_tuple_sz - (K + V);
+        let pad = [0; std::mem::size_of::<u64>()];
+        self.log.write_all(&pad[..pad_sz])?;
+
+        let logged_bytes = 4 + 1 + min_tuple_sz;
 
         self.memtable.insert(k, v);
 
@@ -680,10 +804,6 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
         self.stats.logged_bytes += logged_bytes as u64;
         self.stats.written_bytes += logged_bytes as u64;
 
-        if self.dirty_bytes > self.config.max_log_length {
-            self.flush()?;
-        }
-
         Ok(())
     }
 
@@ -696,6 +816,13 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
     /// been written, fsynced, and the sstable
     /// directory has been fsyced.
     pub fn flush(&mut self) -> Result<()> {
+        #[cfg(test)]
+        {
+            if self.log.tearing {
+                return Ok(());
+            }
+        }
+
         self.log.flush()?;
         self.log.get_mut().sync_all()?;
 
@@ -725,8 +852,6 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
             self.next_sstable_id += 1;
 
-            assert_eq!(self.log.buffer().len(), 0);
-
             let log_file: &mut fs::File = self.log.get_mut();
             log_file.seek(io::SeekFrom::Start(0))?;
             log_file.set_len(0)?;
@@ -762,4 +887,7 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 }
 
 #[cfg(test)]
-mod tests;
+mod tearable;
+
+#[cfg(test)]
+mod fuzz;
